;; Generated package description from bnf-mode.el  -*- no-byte-compile: t -*-
(define-package "bnf-mode" "0.4.5.0.20221205.150230" "Major mode for editing BNF grammars." '((cl-lib "0.5") (emacs "25.1")) :commit "1a7e177c282b8e07a2c33bd89232464b347dfc17" :authors '(("Serghei Iakovlev" . "egrep@protonmail.ch")) :maintainer '("Serghei Iakovlev" . "egrep@protonmail.ch") :keywords '("languages") :url "https://github.com/sergeyklay/bnf-mode")
